@@ -1,5 +1,5 @@
 #!/bin/sh
-test=`basename $0 | sed -e 's|\.x87\.nodata$||'`
+test=`basename $0 | sed -e 's|\.x87.nodata$||'`
 ./lightning -data=0 -mx87=1 $srcdir/$test.tst | tr -d \\r > $test.out
 if test $? != 0; then
   exit $?
