@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012, 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -21,8 +21,6 @@
 #  include <stdio.h>
 #endif
 
-#define rc(value)			jit_class_##value
-#define rn(reg)				(jit_regno(_rvs[jit_regno(reg)].spec))
 #define jit_arg_reg_p(i)		((i) >= 0 && (i) < 4)
 #define jit_arg_f_reg_p(i)		((i) >= 0 && (i) < 16)
 #define jit_arg_d_reg_p(i)		((i) >= 0 && (i) < 15)
@@ -34,6 +32,11 @@
 
 #define jit_fpr_p(rn)			((rn) > 15)
 
+#define arg_base()							\
+    (stack_framesize - 16 + (jit_cpu.abi ? 64 : 0))
+#define arg_offset(n)							\
+    ((n) < 4 ? arg_base() + ((n) << 2) : (n))
+
 /* Assume functions called never match jit instruction set, that is
  * libc, gmp, mpfr, etc functions are in thumb mode and jit is in
  * arm mode, what may cause a crash upon return of that function
@@ -51,9 +54,17 @@ typedef union _jit_thumb_t {
     jit_int16_t		s[2];
 } jit_thumb_t;
 
+typedef jit_pointer_t	jit_va_list;
+
 /*
  * Prototypes
  */
+#define jit_make_arg(node)		_jit_make_arg(_jit,node)
+static jit_node_t *_jit_make_arg(jit_state_t*,jit_node_t*);
+#define jit_make_arg_f(node)		_jit_make_arg_f(_jit,node)
+static jit_node_t *_jit_make_arg_f(jit_state_t*,jit_node_t*);
+#define jit_make_arg_d(node)		_jit_make_arg_d(_jit,node)
+static jit_node_t *_jit_make_arg_d(jit_state_t*,jit_node_t*);
 #define jit_get_reg_pair()		_jit_get_reg_pair(_jit)
 static jit_int32_t _jit_get_reg_pair(jit_state_t*);
 #define jit_unget_reg_pair(rn)		_jit_unget_reg_pair(_jit,rn)
@@ -75,6 +86,7 @@ extern void __clear_cache(void *, void *);
 #endif
 
 #define PROTO				1
+#  include "jit_rewind.c"
 #  include "jit_arm-cpu.c"
 #  include "jit_arm-swf.c"
 #  include "jit_arm-vfp.c"
@@ -245,6 +257,10 @@ _jit_prolog(jit_state_t *_jit)
     jit_alloc((jit_pointer_t *)&_jitc->function->regoff,
 	      _jitc->reglen * sizeof(jit_int32_t));
 
+    /* _no_link here does not mean the jit_link() call can be removed
+     * by rewriting as:
+     * _jitc->function->prolog = jit_new_node(jit_code_prolog);
+     */
     _jitc->function->prolog = jit_new_node_no_link(jit_code_prolog);
     jit_link(_jitc->function->prolog);
     _jitc->function->prolog->w.w = offset;
@@ -269,41 +285,71 @@ _jit_allocai(jit_state_t *_jit, jit_int32_t length)
 	default:	_jitc->function->self.aoff &= -8;	break;
     }
     _jitc->function->self.aoff -= length;
+    if (!_jitc->realize) {
+	jit_inc_synth_ww(allocai, _jitc->function->self.aoff, length);
+	jit_dec_synth();
+    }
     return (_jitc->function->self.aoff);
 }
 
+void
+_jit_allocar(jit_state_t *_jit, jit_int32_t u, jit_int32_t v)
+{
+    jit_int32_t		 reg;
+    assert(_jitc->function);
+    jit_inc_synth_ww(allocar, u, v);
+    if (!_jitc->function->allocar) {
+	_jitc->function->aoffoff = jit_allocai(sizeof(jit_int32_t));
+	_jitc->function->allocar = 1;
+    }
+    reg = jit_get_reg(jit_class_gpr);
+    jit_negr(reg, v);
+    jit_andi(reg, reg, -8);
+    jit_ldxi_i(u, JIT_FP, _jitc->function->aoffoff);
+    jit_addr(u, u, reg);
+    jit_addr(JIT_SP, JIT_SP, reg);
+    jit_stxi_i(_jitc->function->aoffoff, JIT_FP, u);
+    jit_unget_reg(reg);
+    jit_dec_synth();
+}
+
 void
 _jit_ret(jit_state_t *_jit)
 {
     jit_node_t		*instr;
-
     assert(_jitc->function);
-
+    jit_inc_synth(ret);
     /* jump to epilog */
     instr = jit_jmpi();
     jit_patch_at(instr, _jitc->function->epilog);
+    jit_dec_synth();
 }
 
 void
 _jit_retr(jit_state_t *_jit, jit_int32_t u)
 {
+    jit_inc_synth_w(retr, u);
     if (JIT_RET != u)
 	jit_movr(JIT_RET, u);
     else
 	jit_live(JIT_RET);
     jit_ret();
+    jit_dec_synth();
 }
 
 void
 _jit_reti(jit_state_t *_jit, jit_word_t u)
 {
+    jit_inc_synth_w(reti, u);
     jit_movi(JIT_RET, u);
     jit_ret();
+    jit_dec_synth();
 }
 
 void
 _jit_retr_f(jit_state_t *_jit, jit_int32_t u)
 {
+    jit_inc_synth_w(retr_f, u);
     if (jit_cpu.abi) {
 	if (u != JIT_FRET)
 	    jit_movr_f(JIT_FRET, u);
@@ -317,21 +363,25 @@ _jit_retr_f(jit_state_t *_jit, jit_int32_t u)
 	    jit_live(JIT_RET);
     }
     jit_ret();
+    jit_dec_synth();
 }
 
 void
 _jit_reti_f(jit_state_t *_jit, jit_float32_t u)
 {
+    jit_inc_synth_f(reti_f, u);
     if (jit_cpu.abi)
 	jit_movi_f(JIT_FRET, u);
     else
 	jit_movi_f_w(JIT_RET, u);
     jit_ret();
+    jit_dec_synth();
 }
 
 void
 _jit_retr_d(jit_state_t *_jit, jit_int32_t u)
 {
+    jit_inc_synth_w(retr_d, u);
     if (jit_cpu.abi) {
 	if (u != JIT_FRET)
 	    jit_movr_d(JIT_FRET, u);
@@ -345,16 +395,19 @@ _jit_retr_d(jit_state_t *_jit, jit_int32_t u)
 	    jit_live(JIT_RET);
     }
     jit_ret();
+    jit_dec_synth();
 }
 
 void
 _jit_reti_d(jit_state_t *_jit, jit_float64_t u)
 {
+    jit_inc_synth_d(reti_d, u);
     if (jit_cpu.abi)
 	jit_movi_d(JIT_FRET, u);
     else
 	jit_movi_d_ww(JIT_RET, _R1, u);
     jit_ret();
+    jit_dec_synth();
 }
 
 void
@@ -383,25 +436,30 @@ _jit_arg_register_p(jit_state_t *_jit, jit_node_t *u)
     return (jit_arg_reg_p(u->u.w));
 }
 
-jit_node_t *
-_jit_arg(jit_state_t *_jit)
+static jit_node_t *
+_jit_make_arg(jit_state_t *_jit, jit_node_t *node)
 {
-    jit_int32_t		offset;
-    assert(_jitc->function);
+    jit_int32_t		 offset;
     if (jit_arg_reg_p(_jitc->function->self.argi))
 	offset = _jitc->function->self.argi++;
     else {
 	offset = _jitc->function->self.size;
 	_jitc->function->self.size += sizeof(jit_word_t);
     }
-    return (jit_new_node_w(jit_code_arg, offset));
+    if (node == (jit_node_t *)0)
+	node = jit_new_node(jit_code_arg);
+    else
+	link_node(node);
+    node->u.w = offset;
+    node->v.w = ++_jitc->function->self.argn;
+    jit_link_prolog();
+    return (node);
 }
 
 jit_node_t *
-_jit_arg_f(jit_state_t *_jit)
+_jit_make_arg_f(jit_state_t *_jit, jit_node_t *node)
 {
-    jit_int32_t		offset;
-    assert(_jitc->function);
+    jit_int32_t		 offset;
     if (jit_cpu.abi && !(_jitc->function->self.call & jit_call_varargs)) {
 	if (jit_arg_f_reg_p(_jitc->function->self.argf)) {
 	    offset = _jitc->function->self.argf++;
@@ -417,14 +475,20 @@ _jit_arg_f(jit_state_t *_jit)
     offset = _jitc->function->self.size;
     _jitc->function->self.size += sizeof(jit_float32_t);
 done:
-    return (jit_new_node_w(jit_code_arg_f, offset));
+    if (node == (jit_node_t *)0)
+	node = jit_new_node(jit_code_arg_f);
+    else
+	link_node(node);
+    node->u.w = offset;
+    node->v.w = ++_jitc->function->self.argn;
+    jit_link_prolog();
+    return (node);
 }
 
 jit_node_t *
-_jit_arg_d(jit_state_t *_jit)
+_jit_make_arg_d(jit_state_t *_jit, jit_node_t *node)
 {
-    jit_int32_t		offset;
-    assert(_jitc->function);
+    jit_int32_t		 offset;
     if (jit_cpu.abi && !(_jitc->function->self.call & jit_call_varargs)) {
 	if (jit_arg_d_reg_p(_jitc->function->self.argf)) {
 	    if (_jitc->function->self.argf & 1)
@@ -448,79 +512,147 @@ _jit_arg_d(jit_state_t *_jit)
     offset = _jitc->function->self.size;
     _jitc->function->self.size += sizeof(jit_float64_t);
 done:
-    return (jit_new_node_w(jit_code_arg_d, offset));
+    if (node == (jit_node_t *)0)
+	node = jit_new_node(jit_code_arg_d);
+    else
+	link_node(node);
+    node->u.w = offset;
+    node->v.w = ++_jitc->function->self.argn;
+    jit_link_prolog();
+    return (node);
+}
+
+void
+_jit_ellipsis(jit_state_t *_jit)
+{
+    if (_jitc->prepare) {
+	assert(!(_jitc->function->call.call & jit_call_varargs));
+	_jitc->function->call.call |= jit_call_varargs;
+	if (jit_cpu.abi && _jitc->function->call.argf)
+	    rewind_prepare();
+    }
+    else {
+	assert(!(_jitc->function->self.call & jit_call_varargs));
+	_jitc->function->self.call |= jit_call_varargs;
+	if (jit_cpu.abi &&  _jitc->function->self.argf)
+	    rewind_prolog();
+	/* First 4 stack addresses are always spilled r0-r3 */
+	if (jit_arg_reg_p(_jitc->function->self.argi))
+	    _jitc->function->vagp = _jitc->function->self.argi * 4;
+	else
+	    _jitc->function->vagp = 16;
+    }
+    jit_inc_synth(ellipsis);
+    if (_jitc->prepare)
+	jit_link_prepare();
+    else
+	jit_link_prolog();
+    jit_dec_synth();
+}
+
+jit_node_t *
+_jit_arg(jit_state_t *_jit)
+{
+    assert(_jitc->function);
+    return (jit_make_arg((jit_node_t*)0));
+}
+
+jit_node_t *
+_jit_arg_f(jit_state_t *_jit)
+{
+    assert(_jitc->function);
+    return (jit_make_arg_f((jit_node_t*)0));
+}
+
+jit_node_t *
+_jit_arg_d(jit_state_t *_jit)
+{
+    assert(_jitc->function);
+    return (jit_make_arg_d((jit_node_t*)0));
 }
 
 void
 _jit_getarg_c(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(getarg_c, u, v);
     if (jit_swf_p())
-	jit_ldxi_c(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_c(u, JIT_FP, arg_offset(v->u.w));
     else if (jit_arg_reg_p(v->u.w))
 	jit_extr_c(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_c(u, JIT_FP, v->u.w);
+    jit_dec_synth();
 }
 
 void
 _jit_getarg_uc(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(getarg_uc, u, v);
     if (jit_swf_p())
-	jit_ldxi_uc(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_uc(u, JIT_FP, arg_offset(v->u.w));
     else if (jit_arg_reg_p(v->u.w))
 	jit_extr_uc(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_uc(u, JIT_FP, v->u.w);
+    jit_dec_synth();
 }
 
 void
 _jit_getarg_s(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(getarg_s, u, v);
     if (jit_swf_p())
-	jit_ldxi_s(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_s(u, JIT_FP, arg_offset(v->u.w));
     else if (jit_arg_reg_p(v->u.w))
 	jit_extr_s(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_s(u, JIT_FP, v->u.w);
+    jit_dec_synth();
 }
 
 void
 _jit_getarg_us(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(getarg_us, u, v);
     if (jit_swf_p())
-	jit_ldxi_us(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_us(u, JIT_FP, arg_offset(v->u.w));
     else if (jit_arg_reg_p(v->u.w))
 	jit_extr_us(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_us(u, JIT_FP, v->u.w);
+    jit_dec_synth();
 }
 
 void
 _jit_getarg_i(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(getarg_i, u, v);
     if (jit_swf_p())
-	jit_ldxi_i(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_i(u, JIT_FP, arg_offset(v->u.w));
     else if (jit_arg_reg_p(v->u.w))
 	jit_movr(u, JIT_RA0 - v->u.w);
     else
 	jit_ldxi_i(u, JIT_FP, v->u.w);
+    jit_dec_synth();
 }
 
 void
 _jit_putargr(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(putargr, u, v);
     if (jit_swf_p())
-	jit_stxi(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, u);
+	jit_stxi(arg_offset(v->u.w), JIT_FP, u);
     else if (jit_arg_reg_p(v->u.w))
 	jit_movr(JIT_RA0 - v->u.w, u);
     else
 	jit_stxi(v->u.w, JIT_FP, u);
+    jit_dec_synth();
 }
 
 void
@@ -528,10 +660,11 @@ _jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
 {
     jit_int32_t		regno;
     assert(v->code == jit_code_arg);
+    jit_inc_synth_wp(putargi, u, v);
     if (jit_swf_p()) {
 	regno = jit_get_reg(jit_class_gpr);
 	jit_movi(regno, u);
-	jit_stxi(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, regno);
+	jit_stxi(arg_offset(v->u.w), JIT_FP, regno);
 	jit_unget_reg(regno);
     }
     else if (jit_arg_reg_p(v->u.w))
@@ -542,32 +675,36 @@ _jit_putargi(jit_state_t *_jit, jit_word_t u, jit_node_t *v)
 	jit_stxi(v->u.w, JIT_FP, regno);
 	jit_unget_reg(regno);
     }
+    jit_dec_synth();
 }
 
 void
 _jit_getarg_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg_f);
-    if (jit_cpu.abi) {
+    jit_inc_synth_wp(getarg_f, u, v);
+    if (jit_cpu.abi && !(_jitc->function->self.call & jit_call_varargs)) {
 	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movr_f(u, JIT_FA0 - v->u.w);
 	else
 	    jit_ldxi_f(u, JIT_FP, v->u.w);
     }
     else if (jit_swf_p())
-	jit_ldxi_f(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_f(u, JIT_FP, arg_offset(v->u.w));
     else {
 	if (jit_arg_reg_p(v->u.w))
 	    jit_movr_w_f(u, JIT_RA0 - v->u.w);
 	else
 	    jit_ldxi_f(u, JIT_FP, v->u.w);
     }
+    jit_dec_synth();
 }
 
 void
 _jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg_f);
+    jit_inc_synth_wp(putargr_f, u, v);
     if (jit_cpu.abi) {
 	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movr_f(JIT_FA0 - v->u.w, u);
@@ -575,13 +712,14 @@ _jit_putargr_f(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	    jit_stxi_f(v->u.w, JIT_FP, u);
     }
     else if (jit_swf_p())
-	jit_stxi_f(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, u);
+	jit_stxi_f(arg_offset(v->u.w), JIT_FP, u);
     else {
 	if (jit_arg_reg_p(v->u.w))
 	    jit_movr_f_w(JIT_RA0 - v->u.w, u);
 	else
 	    jit_stxi_f(v->u.w, JIT_FP, u);
     }
+    jit_dec_synth();
 }
 
 void
@@ -589,6 +727,7 @@ _jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
 {
     jit_int32_t		regno;
     assert(v->code == jit_code_arg_f);
+    jit_inc_synth_fp(putargi_f, u, v);
     if (jit_cpu.abi) {
 	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movi_f(JIT_FA0 - v->u.w, u);
@@ -602,7 +741,7 @@ _jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
     else if (jit_swf_p()) {
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_f(regno, u);
-	jit_stxi_f(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, regno);
+	jit_stxi_f(arg_offset(v->u.w), JIT_FP, regno);
 	jit_unget_reg(regno);
     }
     else {
@@ -614,32 +753,36 @@ _jit_putargi_f(jit_state_t *_jit, jit_float32_t u, jit_node_t *v)
 	    jit_stxi_f(v->u.w, JIT_FP, regno);
 	jit_unget_reg(regno);
     }
+    jit_dec_synth();
 }
 
 void
 _jit_getarg_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg_d);
-    if (jit_cpu.abi) {
+    jit_inc_synth_wp(getarg_d, u, v);
+    if (jit_cpu.abi && !(_jitc->function->self.call & jit_call_varargs)) {
 	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movr_d(u, JIT_FA0 - v->u.w);
 	else
 	    jit_ldxi_d(u, JIT_FP, v->u.w);
     }
     else if (jit_swf_p())
-	jit_ldxi_d(u, JIT_FP, v->u.w < 4 ? v->u.w << 2 : v->u.w);
+	jit_ldxi_d(u, JIT_FP, arg_offset(v->u.w));
     else {
 	if (jit_arg_reg_p(v->u.w))
 	    jit_movr_ww_d(u, JIT_RA0 - v->u.w, JIT_RA0 - (v->u.w + 1));
 	else
 	    jit_ldxi_d(u, JIT_FP, v->u.w);
     }
+    jit_dec_synth();
 }
 
 void
 _jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 {
     assert(v->code == jit_code_arg_d);
+    jit_inc_synth_wp(putargr_d, u, v);
     if (jit_cpu.abi) {
 	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movr_d(JIT_FA0 - v->u.w, u);
@@ -647,13 +790,14 @@ _jit_putargr_d(jit_state_t *_jit, jit_int32_t u, jit_node_t *v)
 	    jit_stxi_d(v->u.w, JIT_FP, u);
     }
     else if (jit_swf_p())
-	jit_stxi_d(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, u);
+	jit_stxi_d(arg_offset(v->u.w), JIT_FP, u);
     else {
 	if (jit_arg_reg_p(v->u.w))
 	    jit_movr_d_ww(JIT_RA0 - v->u.w, JIT_RA0 - (v->u.w + 1), u);
 	else
 	    jit_stxi_d(v->u.w, JIT_FP, u);
     }
+    jit_dec_synth();
 }
 
 void
@@ -661,6 +805,7 @@ _jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
 {
     jit_int32_t		regno;
     assert(v->code == jit_code_arg_d);
+    jit_inc_synth_dp(putargi_d, u, v);
     if (jit_cpu.abi) {
 	if (jit_arg_f_reg_p(v->u.w))
 	    jit_movi_d(JIT_FA0 - v->u.w, u);
@@ -674,7 +819,7 @@ _jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
     else if (jit_swf_p()) {
 	regno = jit_get_reg(jit_class_fpr);
 	jit_movi_d(regno, u);
-	jit_stxi_d(v->u.w < 4 ? v->u.w << 2 : v->u.w, JIT_FP, regno);
+	jit_stxi_d(arg_offset(v->u.w), JIT_FP, regno);
 	jit_unget_reg(regno);
     }
     else {
@@ -686,12 +831,15 @@ _jit_putargi_d(jit_state_t *_jit, jit_float64_t u, jit_node_t *v)
 	    jit_stxi_d(v->u.w, JIT_FP, regno);
 	jit_unget_reg(regno);
     }
+    jit_dec_synth();
 }
 
 void
 _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
+    jit_inc_synth_w(pushargr, u);
+    jit_link_prepare();
     if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movr(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
@@ -700,6 +848,7 @@ _jit_pushargr(jit_state_t *_jit, jit_int32_t u)
 	jit_stxi(_jitc->function->call.size, JIT_SP, u);
 	_jitc->function->call.size += sizeof(jit_word_t);
     }
+    jit_dec_synth();
 }
 
 void
@@ -707,6 +856,8 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 {
     jit_int32_t		 regno;
     assert(_jitc->function);
+    jit_inc_synth_w(pushargi, u);
+    jit_link_prepare();
     if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	jit_movi(JIT_RA0 - _jitc->function->call.argi, u);
 	++_jitc->function->call.argi;
@@ -718,28 +869,33 @@ _jit_pushargi(jit_state_t *_jit, jit_word_t u)
 	jit_unget_reg(regno);
 	_jitc->function->call.size += sizeof(jit_word_t);
     }
+    jit_dec_synth();
 }
 
 void
 _jit_pushargr_f(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
+    jit_inc_synth_w(pushargr_f, u);
+    jit_link_prepare();
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
 	if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	    jit_movr_f(JIT_FA0 - _jitc->function->call.argf, u);
 	    ++_jitc->function->call.argf;
-	    return;
+	    goto done;
 	}
     }
     else {
 	if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	    jit_movr_f_w(JIT_RA0 - _jitc->function->call.argi, u);
 	    ++_jitc->function->call.argi;
-	    return;
+	    goto done;
 	}
     }
     jit_stxi_f(_jitc->function->call.size, JIT_SP, u);
     _jitc->function->call.size += sizeof(jit_word_t);
+done:
+    jit_dec_synth();
 }
 
 void
@@ -747,6 +903,8 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 {
     jit_int32_t		regno;
     assert(_jitc->function);
+    jit_inc_synth_f(pushargi_f, u);
+    jit_link_prepare();
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
 	if (jit_arg_f_reg_p(_jitc->function->call.argf)) {
 	    /* cannot jit_movi_f in the argument register because
@@ -758,14 +916,14 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
 	    jit_movr_f(JIT_FA0 - _jitc->function->call.argf, regno);
 	    jit_unget_reg(regno);
 	    ++_jitc->function->call.argf;
-	    return;
+	    goto done;
 	}
     }
     else {
 	if (jit_arg_reg_p(_jitc->function->call.argi)) {
 	    jit_movi_f_w(JIT_RA0 - _jitc->function->call.argi, u);
 	    ++_jitc->function->call.argi;
-	    return;
+	    goto done;
 	}
     }
     regno = jit_get_reg(jit_class_fpr);
@@ -773,19 +931,23 @@ _jit_pushargi_f(jit_state_t *_jit, jit_float32_t u)
     jit_stxi_f(_jitc->function->call.size, JIT_SP, regno);
     jit_unget_reg(regno);
     _jitc->function->call.size += sizeof(jit_word_t);
+done:
+    jit_dec_synth();
 }
 
 void
 _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 {
     assert(_jitc->function);
+    jit_inc_synth_w(pushargr_d, u);
+    jit_link_prepare();
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
 	if (jit_arg_d_reg_p(_jitc->function->call.argf)) {
 	    if (_jitc->function->call.argf & 1)
 		++_jitc->function->call.argf;
 	    jit_movr_d(JIT_FA0 - _jitc->function->call.argf, u);
 	    _jitc->function->call.argf += 2;
-	    return;
+	    goto done;
 	}
     }
     else {
@@ -796,13 +958,15 @@ _jit_pushargr_d(jit_state_t *_jit, jit_int32_t u)
 			  JIT_RA0 - (_jitc->function->call.argi + 1),
 			  u);
 	    _jitc->function->call.argi += 2;
-	    return;
+	    goto done;
 	}
     }
     if (_jitc->function->call.size & 7)
 	_jitc->function->call.size += 4;
     jit_stxi_d(_jitc->function->call.size, JIT_SP, u);
     _jitc->function->call.size += sizeof(jit_float64_t);
+done:
+    jit_dec_synth();
 }
 
 void
@@ -810,13 +974,15 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
 {
     jit_int32_t		regno;
     assert(_jitc->function);
+    jit_inc_synth_d(pushargi_d, u);
+    jit_link_prepare();
     if (jit_cpu.abi && !(_jitc->function->call.call & jit_call_varargs)) {
 	if (jit_arg_d_reg_p(_jitc->function->call.argf)) {
 	    if (_jitc->function->call.argf & 1)
 		++_jitc->function->call.argf;
 	    jit_movi_d(JIT_FA0 - _jitc->function->call.argf, u);
 	    _jitc->function->call.argf += 2;
-	    return;
+	    goto done;
 	}
     }
     else {
@@ -827,7 +993,7 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
 			  JIT_RA0 - (_jitc->function->call.argi + 1),
 			  u);
 	    _jitc->function->call.argi += 2;
-	    return;
+	    goto done;
 	}
     }
     if (_jitc->function->call.size & 7)
@@ -837,6 +1003,8 @@ _jit_pushargi_d(jit_state_t *_jit, jit_float64_t u)
     jit_stxi_d(_jitc->function->call.size, JIT_SP, regno);
     jit_unget_reg(regno);
     _jitc->function->call.size += sizeof(jit_float64_t);
+done:
+    jit_dec_synth();
 }
 
 jit_bool_t
@@ -863,6 +1031,7 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
 {
     jit_node_t		*node;
     assert(_jitc->function);
+    jit_inc_synth_w(finishr, r0);
     if (_jitc->function->self.alen < _jitc->function->call.size)
 	_jitc->function->self.alen = _jitc->function->call.size;
     node = jit_callr(r0);
@@ -871,6 +1040,7 @@ _jit_finishr(jit_state_t *_jit, jit_int32_t r0)
     _jitc->function->call.argi = _jitc->function->call.argf =
 	_jitc->function->call.size = 0;
     _jitc->prepare = 0;
+    jit_dec_synth();
 }
 
 jit_node_t *
@@ -878,6 +1048,7 @@ _jit_finishi(jit_state_t *_jit, jit_pointer_t i0)
 {
     jit_node_t		*node;
     assert(_jitc->function);
+    jit_inc_synth_w(finishi, (jit_word_t)i0);
     if (_jitc->function->self.alen < _jitc->function->call.size)
 	_jitc->function->self.alen = _jitc->function->call.size;
     node = jit_calli(i0);
@@ -886,60 +1057,75 @@ _jit_finishi(jit_state_t *_jit, jit_pointer_t i0)
     _jitc->function->call.argi = _jitc->function->call.argf =
 	_jitc->function->call.size = 0;
     _jitc->prepare = 0;
+    jit_dec_synth();
     return (node);
 }
 
 void
 _jit_retval_c(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_c, r0);
     jit_extr_c(r0, JIT_RET);
+    jit_dec_synth();
 }
 
 void
 _jit_retval_uc(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_uc, r0);
     jit_extr_uc(r0, JIT_RET);
+    jit_dec_synth();
 }
 
 void
 _jit_retval_s(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_s, r0);
     jit_extr_s(r0, JIT_RET);
+    jit_dec_synth();
 }
 
 void
 _jit_retval_us(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_us, r0);
     jit_extr_us(r0, JIT_RET);
+    jit_dec_synth();
 }
 
 void
 _jit_retval_i(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_i, r0);
     if (r0 != JIT_RET)
 	jit_movr(r0, JIT_RET);
+    jit_dec_synth();
 }
 
 void
 _jit_retval_f(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_f, r0);
     if (jit_cpu.abi) {
 	if (r0 != JIT_FRET)
 	    jit_movr_f(r0, JIT_FRET);
     }
     else if (r0 != JIT_RET)
 	jit_movr_w_f(r0, JIT_RET);
+    jit_dec_synth();
 }
 
 void
 _jit_retval_d(jit_state_t *_jit, jit_int32_t r0)
 {
+    jit_inc_synth_w(retval_d, r0);
     if (jit_cpu.abi) {
 	if (r0 != JIT_FRET)
 	    jit_movr_d(r0, JIT_FRET);
     }
     else if (r0 != JIT_RET)
 	jit_movr_ww_d(r0, JIT_RET, _R1);
+    jit_dec_synth();
 }
 
 jit_pointer_t
@@ -954,6 +1140,9 @@ _emit_code(jit_state_t *_jit)
 	jit_node_t	*node;
 	jit_uint8_t	*data;
 	jit_word_t	 word;
+#if DEVEL_DISASSEMBLER
+	jit_word_t	 prevw;
+#endif
 	jit_uword_t	 thumb;
 #if DISASSEMBLER
 	jit_int32_t	 info_offset;
@@ -961,12 +1150,18 @@ _emit_code(jit_state_t *_jit)
 	jit_int32_t	 const_offset;
 	jit_int32_t	 patch_offset;
     } undo;
+#if DEVEL_DISASSEMBLER
+    jit_word_t		 prevw;
+#endif
 
     _jitc->function = NULL;
     _jitc->thumb = 0;
 
     jit_reglive_setup();
 
+    _jitc->consts.data = NULL;
+    _jitc->consts.offset = _jitc->consts.length = 0;
+
     undo.word = 0;
     undo.node = NULL;
     undo.data = NULL;
@@ -1182,12 +1377,16 @@ _emit_code(jit_state_t *_jit)
 		    patch(word, node);					\
 		}							\
 		break
+#if DEVEL_DISASSEMBLER
+    prevw = _jit->pc.w;
+#endif
     for (node = _jitc->head; node; node = node->next) {
 	if (_jit->pc.uc >= _jitc->code.end)
 	    return (NULL);
 
 #if DEVEL_DISASSEMBLER
-	node->offset = _jit->pc.w;
+	node->offset = (jit_uword_t)_jit->pc.w - (jit_uword_t)prevw;
+	prevw = _jit->pc.w;
 #endif
 	value = jit_classify(node->code);
 	jit_regarg_set(node, value);
@@ -1588,6 +1787,9 @@ _emit_code(jit_state_t *_jit)
 		_jitc->function = _jitc->functions.ptr + node->w.w;
 		undo.node = node;
 		undo.word = _jit->pc.w;
+#if DEVEL_DISASSEMBLER
+		undo.prevw = prevw;
+#endif
 		undo.data = _jitc->consts.data;
 		undo.thumb = _jitc->thumb;
 		undo.const_offset = _jitc->consts.offset;
@@ -1612,6 +1814,9 @@ _emit_code(jit_state_t *_jit)
 		    temp->flag &= ~jit_flag_patch;
 		    node = undo.node;
 		    _jit->pc.w = undo.word;
+#if DEVEL_DISASSEMBLER
+		    prevw = undo.prevw;
+#endif
 		    invalidate_consts();
 		    _jitc->consts.data = undo.data;
 		    _jitc->thumb = undo.thumb;
@@ -1668,15 +1873,49 @@ _emit_code(jit_state_t *_jit)
 		else
 		    vfp_movi_d(rn(node->u.w), node->w.d);
 		break;
-	    case jit_code_live:
+	    case jit_code_va_start:
+		vastart(rn(node->u.w));
+		break;
+	    case jit_code_va_arg:
+		vaarg(rn(node->u.w), rn(node->v.w));
+		break;
+	    case jit_code_va_arg_d:
+		if (jit_swf_p())
+		    swf_vaarg_d(rn(node->u.w), rn(node->v.w));
+		else
+		    vfp_vaarg_d(rn(node->u.w), rn(node->v.w));
+		break;
+	    case jit_code_live:			case jit_code_ellipsis:
+	    case jit_code_allocai:		case jit_code_allocar:
 	    case jit_code_arg:
 	    case jit_code_arg_f:		case jit_code_arg_d:
+	    case jit_code_va_end:
+	    case jit_code_ret:
+	    case jit_code_retr:			case jit_code_reti:
+	    case jit_code_retr_f:		case jit_code_reti_f:
+	    case jit_code_retr_d:		case jit_code_reti_d:
+	    case jit_code_getarg_c:		case jit_code_getarg_uc:
+	    case jit_code_getarg_s:		case jit_code_getarg_us:
+	    case jit_code_getarg_i:
+	    case jit_code_getarg_f:		case jit_code_getarg_d:
+	    case jit_code_putargr:		case jit_code_putargi:
+	    case jit_code_putargr_f:		case jit_code_putargi_f:
+	    case jit_code_putargr_d:		case jit_code_putargi_d:
+	    case jit_code_pushargr:		case jit_code_pushargi:
+	    case jit_code_pushargr_f:		case jit_code_pushargi_f:
+	    case jit_code_pushargr_d:		case jit_code_pushargi_d:
+	    case jit_code_retval_c:		case jit_code_retval_uc:
+	    case jit_code_retval_s:		case jit_code_retval_us:
+	    case jit_code_retval_i:
+	    case jit_code_retval_f:		case jit_code_retval_d:
+	    case jit_code_prepare:
+	    case jit_code_finishr:		case jit_code_finishi:
 		break;
 	    default:
 		abort();
 	}
 	jit_regarg_clr(node, value);
-	assert(_jitc->regarg == 0);
+	assert(_jitc->regarg == 0 && _jitc->synth == 0);
 	/* update register live state */
 	jit_reglive(node);
 
@@ -1748,6 +1987,7 @@ _emit_code(jit_state_t *_jit)
 }
 
 #define CODE				1
+#  include "jit_rewind.c"
 #  include "jit_arm-cpu.c"
 #  include "jit_arm-swf.c"
 #  include "jit_arm-vfp.c"
