@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012, 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -73,8 +73,6 @@ jit_init_debug(const char *progname)
     }
     bfd_check_format(disasm_bfd, bfd_object);
     bfd_check_format(disasm_bfd, bfd_archive);
-    disasm_print = disassembler(disasm_bfd);
-    assert(disasm_print);
     INIT_DISASSEMBLE_INFO(disasm_info, disasm_stream, fprintf);
 #  if defined(__i386__) || defined(__x86_64__)
     disasm_info.arch = bfd_arch_i386;
@@ -124,6 +122,14 @@ jit_init_debug(const char *progname)
 #  endif
     disasm_info.print_address_func = disasm_print_address;
 
+# if BINUTILS_2_29
+    disasm_print = disassembler(disasm_info.arch, __BYTE_ORDER == __BIG_ENDIAN,
+				disasm_info.mach, disasm_bfd);
+#  else
+    disasm_print = disassembler(disasm_bfd);
+#  endif
+    assert(disasm_print);
+
     if (bfd_get_file_flags(disasm_bfd) & HAS_SYMS) {
 	asymbol		**in;
 	asymbol		**out;
@@ -329,6 +335,7 @@ _disassemble(jit_state_t *_jit, jit_pointer_t code, jit_int32_t length)
     char		 buffer[address_buffer_length];
 #if DEVEL_DISASSEMBLER
     jit_node_t		*node;
+    jit_uword_t		 prevw;
 #endif
 
 #if __arm__
@@ -343,14 +350,18 @@ _disassemble(jit_state_t *_jit, jit_pointer_t code, jit_int32_t length)
     disasm_jit = _jit;
 #if DEVEL_DISASSEMBLER
     node = _jitc->head;
+    prevw = pc;
 #endif
     while (pc < end) {
 #if DEVEL_DISASSEMBLER
-	while (node && node->offset < (jit_uword_t)pc)
+	while (node && (jit_uword_t)(prevw + node->offset) < (jit_uword_t)pc) {
+	    prevw += node->offset;
 	    node = node->next;
-	while (node && node->offset == (jit_uword_t)pc) {
+	}
+	while (node && (jit_uword_t)(prevw + node->offset) == (jit_uword_t)pc) {
 	    jit_print_node(node);
 	    fputc('\n', stdout); 
+	    prevw += node->offset;
 	    node = node->next;
 	}
 #endif
