@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2013-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -435,14 +435,18 @@ static void F16_(jit_state_t*,jit_word_t,
 static void _movi_f(jit_state_t*,jit_int32_t,jit_float32_t*);
 #define movi_d(r0,i0)			_movi_d(_jit,r0,i0)
 static void _movi_d(jit_state_t*,jit_int32_t,jit_float64_t*);
+#define movr_w_f(r0,r1)			_movr_w_f(_jit,r0,r1)
+static void _movr_w_f(jit_state_t*,jit_int32_t,jit_int32_t);
 #define movr_f_w(r0,r1)			_movr_f_w(_jit,r0,r1)
 static void _movr_f_w(jit_state_t*,jit_int32_t,jit_int32_t);
+#define movr_w_d(r0,r1)			_movr_w_d(_jit,r0,r1)
+static void _movr_w_d(jit_state_t*,jit_int32_t,jit_int32_t);
 #define movr_d_w(r0,r1)			_movr_d_w(_jit,r0,r1)
 static void _movr_d_w(jit_state_t*,jit_int32_t,jit_int32_t);
 #define movi_f_w(r0,i0)			_movi_f_w(_jit,r0,i0)
-static void _movi_f_w(jit_state_t*,jit_int32_t,jit_word_t);
+static void _movi_f_w(jit_state_t*,jit_int32_t,jit_float32_t*);
 #define movi_d_w(r0,i0)			_movi_d_w(_jit,r0,i0)
-static void _movi_d_w(jit_state_t*,jit_int32_t,jit_word_t);
+static void _movi_d_w(jit_state_t*,jit_int32_t,jit_float64_t*);
 #define absr_f(r0,r1)			absr_d(r0,r1)
 #define absr_d(r0,r1)			FABS(r0,r1)
 #define negr_f(r0,r1)			negr_d(r0,r1)
@@ -718,6 +722,8 @@ static jit_word_t _bunordr_d(jit_state_t*,jit_word_t,jit_int32_t,jit_int32_t);
 static jit_word_t _bunordi_f(jit_state_t*,jit_word_t,jit_int32_t,jit_float32_t*);
 #define bunordi_d(i0,r0,i1)		_bunordi_d(_jit,i0,r0,i1)
 static jit_word_t _bunordi_d(jit_state_t*,jit_word_t,jit_int32_t,jit_float64_t*);
+#  define vaarg_d(r0, r1)		_vaarg_d(_jit, r0, r1)
+static void _vaarg_d(jit_state_t*, jit_int32_t, jit_int32_t);
 #endif
 
 #if CODE
@@ -1047,6 +1053,15 @@ _movi_d(jit_state_t *_jit, jit_int32_t r0, jit_float64_t *i0)
 	ldi_d(r0, (jit_word_t)i0);
 }
 
+static void
+_movr_w_f(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    /* Should be used only in this case (with out0 == 120) */
+    if (r1 >= 120)
+	r1 = _jitc->rout + (r1 - 120);
+    SETF_S(r0, r1);
+}
+
 static void
 _movr_f_w(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
@@ -1057,30 +1072,56 @@ _movr_f_w(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 }
 
 static void
-_movi_f_w(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
+_movi_f_w(jit_state_t *_jit, jit_int32_t r0, jit_float32_t *i0)
 {
-    /* Should actually be used only in this case (with out0 == 120) */
+    jit_data_t		data;
+
+    /* Should be used only in this case (with out0 == 120) */
     if (r0 >= 120)
 	r0 = _jitc->rout + (r0 - 120);
-    ldi_i(r0, i0);
+    if (_jitc->no_data) {
+	data.f = *i0;
+	movi(r0, data.q.l);
+    }
+    else
+	ldi_i(r0, (jit_word_t)i0);
+}
+
+static void
+_movr_w_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    /* Should be used only in this case (with out0 == 120) */
+    if (r1 >= 120)
+	r1 = _jitc->rout + (r1 - 120);
+    SETF_D(r0, r1);
 }
 
 static void
 _movr_d_w(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
-    /* Should actually be used only in this case (with out0 == 120) */
+    /* Should be used only in this case (with out0 == 120) */
     if (r0 >= 120)
 	r0 = _jitc->rout + (r0 - 120);
     GETF_D(r0, r1);
 }
 
 static void
-_movi_d_w(jit_state_t *_jit, jit_int32_t r0, jit_word_t i0)
+_movi_d_w(jit_state_t *_jit, jit_int32_t r0, jit_float64_t *i0)
 {
-    /* Should actually be used only in this case (with out0 == 120) */
+    union {
+	jit_word_t	 w;
+	jit_float64_t	 d;
+    } data;
+
+    /* Should be used only in this case (with out0 == 120) */
     if (r0 >= 120)
 	r0 = _jitc->rout + (r0 - 120);
-    ldi_l(r0, i0);
+    if (_jitc->no_data) {
+	data.d = *i0;
+	movi(r0, data.w);
+    }
+    else
+	ldi_l(r0, (jit_word_t)i0);
 }
 
 #define fpr_opi(name, type, size)					\
@@ -1708,4 +1749,14 @@ _bunordr_d(jit_state_t *_jit, jit_word_t i0, jit_int32_t r0, jit_int32_t r1)
 }
 fbopi(unord)
 dbopi(unord)
+
+static void
+_vaarg_d(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    assert(_jitc->function->self.call & jit_call_varargs);
+    /* Load argument. */
+    ldr_d(r0, r1);
+    /* Update va_list. */
+    addi(r1, r1, 8);
+}
 #endif
