@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012, 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -92,12 +92,14 @@ typedef union {
 #    else
 #      define stack_framesize		112
 #    endif
+#    define ldr(u,v)			ldr_i(u,v)
 #    define ldi(u,v)			ldi_i(u,v)
 #    define ldxi(u,v,w)			ldxi_i(u,v,w)
 #    define sti(u,v)			sti_i(u,v)
 #    define stxi(u,v,w)			stxi_i(u,v,w)
 #  else
 #    define stack_framesize		144
+#    define ldr(u,v)			ldr_l(u,v)
 #    define ldi(u,v)			ldi_l(u,v)
 #    define ldxi(u,v,w)			ldxi_l(u,v,w)
 #    define sti(u,v)			sti_l(u,v)
@@ -737,6 +739,10 @@ static jit_word_t _calli_p(jit_state_t*,jit_word_t);
 static void _prolog(jit_state_t*,jit_node_t*);
 #  define epilog(node)			_epilog(_jit,node)
 static void _epilog(jit_state_t*,jit_node_t*);
+#  define vastart(r0)			_vastart(_jit, r0)
+static void _vastart(jit_state_t*, jit_int32_t);
+#  define vaarg(r0, r1)			_vaarg(_jit, r0, r1)
+static void _vaarg(jit_state_t*, jit_int32_t, jit_int32_t);
 #define patch_abs(instr,label)		_patch_abs(_jit,instr,label)
 static void _patch_abs(jit_state_t*,jit_word_t,jit_word_t);
 #define patch_at(jump,label)		_patch_at(_jit,jump,label)
@@ -1710,7 +1716,7 @@ _htonr_ui(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 }
 
 static void
-_htonr(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+_htonr_ul(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
 {
     jit_int32_t		reg;
     reg = jit_get_reg(jit_class_gpr);
@@ -2909,6 +2915,8 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 	    return;
 	_jitc->function->self.aoff = frame;
     }
+    if (_jitc->function->allocar)
+	_jitc->function->self.aoff &= -8;
 #if NEW_ABI
     _jitc->function->stack = ((_jitc->function->self.alen -
 			      /* align stack at 16 bytes */
@@ -2921,7 +2929,13 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
 			      _jitc->function->self.aoff) + 7) & -8;
 #endif
     /* callee save registers */
-    subi(_SP_REGNO, _SP_REGNO, stack_framesize);
+#if NEW_ABI
+    if ((_jitc->function->self.call & jit_call_varargs) &&
+	jit_arg_reg_p(_jitc->function->vagp))
+	subi(_SP_REGNO, _SP_REGNO, stack_framesize + 64);
+    else
+#endif
+	subi(_SP_REGNO, _SP_REGNO, stack_framesize);
     offset = stack_framesize - (sizeof(jit_word_t) << 1);
     for (index = 0; index < jit_size(fregs); index++, offset -= 8) {
 	if (jit_regset_tstbit(&_jitc->function->regset, fregs[index]))
@@ -2940,6 +2954,23 @@ _prolog(jit_state_t *_jit, jit_node_t *node)
     /* alloca */
     if (_jitc->function->stack)
 	subi(_SP_REGNO, _SP_REGNO, _jitc->function->stack);
+    if (_jitc->function->allocar) {
+	index = jit_get_reg(jit_class_gpr);
+	movi(rn(index), _jitc->function->self.aoff);
+	stxi_i(_jitc->function->aoffoff, _BP_REGNO, rn(index));
+	jit_unget_reg(index);
+    }
+
+    if (_jitc->function->self.call & jit_call_varargs) {
+#if NEW_ABI
+	index = _jitc->function->vagp;
+#else
+	index = (_jitc->function->self.size - stack_framesize) >> STACK_SHIFT;
+#endif
+	offset = stack_framesize + index * STACK_SLOT;
+	for (; jit_arg_reg_p(index); ++index, offset += STACK_SLOT)
+	    stxi(offset	+  WORD_ADJUST, _BP_REGNO, rn(_A0 - index));
+    }
 }
 
 static void
@@ -2966,7 +2997,41 @@ _epilog(jit_state_t *_jit, jit_node_t *node)
     ldxi(_BP_REGNO, _SP_REGNO, 0);
     JR(_RA_REGNO);
     /* delay slot */
-    addi(_SP_REGNO, _SP_REGNO, stack_framesize);
+#if NEW_ABI
+    if ((_jitc->function->self.call & jit_call_varargs) &&
+	jit_arg_reg_p(_jitc->function->vagp))
+	addi(_SP_REGNO, _SP_REGNO, stack_framesize + 64);
+    else
+#endif
+	addi(_SP_REGNO, _SP_REGNO, stack_framesize);
+}
+
+static void
+_vastart(jit_state_t *_jit, jit_int32_t r0)
+{
+    assert(_jitc->function->self.call & jit_call_varargs);
+    /* Initialize va_list to the first stack argument. */
+#if NEW_ABI
+    if (jit_arg_reg_p(_jitc->function->vagp))
+	addi(r0, _BP_REGNO, stack_framesize + _jitc->function->vagp *
+	     sizeof(jit_word_t));
+    else
+#endif
+	addi(r0, _BP_REGNO, _jitc->function->self.size);
+}
+
+static void
+_vaarg(jit_state_t *_jit, jit_int32_t r0, jit_int32_t r1)
+{
+    /* Load argument. */
+#if WORD_ADJUST
+    ldxi(r0, r1, WORD_ADJUST);
+#else
+    ldr(r0, r1);
+#endif
+
+    /* Update va_list. */
+    addi(r1, r1, STACK_SLOT);
 }
 
 static void
