@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2013-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -26,7 +26,7 @@
 /*
  * Initialization
  */
-static jit_int16_t	_szs[jit_code_last_code + 1] = {
+static jit_int16_t	_szs[jit_code_last_code] = {
 #if GET_JIT_SIZE
 #  define JIT_INSTR_MAX		256
 #else
