@@ -9,8 +9,24 @@
     0,	/* #name */
     0,	/* #note */
     3,	/* label */
-    26,	/* prolog */
+    34,	/* prolog */
+    0,	/* ellipsis */
+    0,	/* allocai */
+    0,	/* allocar */
     0,	/* arg */
+    0,	/* getarg_c */
+    0,	/* getarg_uc */
+    0,	/* getarg_s */
+    0,	/* getarg_us */
+    0,	/* getarg_i */
+    0,	/* getarg_ui */
+    0,	/* getarg_l */
+    0,	/* putargr */
+    0,	/* putargi */
+    3,	/* va_start */
+    5,	/* va_arg */
+    7,	/* va_arg_d */
+    0,	/* va_end */
     3,	/* addr */
     6,	/* addi */
     4,	/* addcr */
@@ -23,7 +39,7 @@
     6,	/* subci */
     6,	/* subxr */
     5,	/* subxi */
-    10,	/* rsbi */
+    8,	/* rsbi */
     5,	/* mulr */
     7,	/* muli */
     20,	/* qmulr */
@@ -84,7 +100,7 @@
     3,	/* extr_us */
     0,	/* extr_i */
     0,	/* extr_ui */
-    4,	/* htonr_us */
+    7,	/* htonr_us */
     4,	/* htonr_ui */
     0,	/* htonr_ul */
     3,	/* ldr_c */
@@ -102,13 +118,13 @@
     0,	/* ldr_l */
     0,	/* ldi_l */
     4,	/* ldxr_c */
-    4,	/* ldxi_c */
+    7,	/* ldxi_c */
     4,	/* ldxr_uc */
-    4,	/* ldxi_uc */
+    7,	/* ldxi_uc */
     4,	/* ldxr_s */
-    4,	/* ldxi_s */
+    7,	/* ldxi_s */
     4,	/* ldxr_us */
-    4,	/* ldxi_us */
+    7,	/* ldxi_us */
     3,	/* ldxr_i */
     6,	/* ldxi_i */
     0,	/* ldxr_ui */
@@ -126,7 +142,7 @@
     11,	/* stxr_c */
     11,	/* stxi_c */
     4,	/* stxr_s */
-    4,	/* stxi_s */
+    7,	/* stxi_s */
     3,	/* stxr_i */
     6,	/* stxi_i */
     0,	/* stxr_l */
@@ -173,15 +189,33 @@
     9,	/* bxsubi_u */
     2,	/* jmpr */
     5,	/* jmpi */
-    4,	/* callr */
-    7,	/* calli */
+    2,	/* callr */
+    5,	/* calli */
+    0,	/* prepare */
+    0,	/* pushargr */
+    0,	/* pushargi */
+    0,	/* finishr */
+    0,	/* finishi */
+    0,	/* ret */
+    0,	/* retr */
+    0,	/* reti */
+    0,	/* retval_c */
+    0,	/* retval_uc */
+    0,	/* retval_s */
+    0,	/* retval_us */
+    0,	/* retval_i */
+    0,	/* retval_ui */
+    0,	/* retval_l */
     24,	/* epilog */
     0,	/* arg_f */
+    0,	/* getarg_f */
+    0,	/* putargr_f */
+    0,	/* putargi_f */
     8,	/* addr_f */
     19,	/* addi_f */
     12,	/* subr_f */
     19,	/* subi_f */
-    19,	/* rsbi_f */
+    21,	/* rsbi_f */
     8,	/* mulr_f */
     19,	/* muli_f */
     12,	/* divr_f */
@@ -226,11 +260,11 @@
     4,	/* ldr_f */
     8,	/* ldi_f */
     5,	/* ldxr_f */
-    5,	/* ldxi_f */
+    8,	/* ldxi_f */
     6,	/* str_f */
     10,	/* sti_f */
     7,	/* stxr_f */
-    7,	/* stxi_f */
+    8,	/* stxi_f */
     10,	/* bltr_f */
     23,	/* blti_f */
     10,	/* bler_f */
@@ -259,12 +293,20 @@
     23,	/* bordi_f */
     10,	/* bunordr_f */
     23,	/* bunordi_f */
+    0,	/* pushargr_f */
+    0,	/* pushargi_f */
+    0,	/* retr_f */
+    0,	/* reti_f */
+    10,	/* retval_f */
     0,	/* arg_d */
+    0,	/* getarg_d */
+    0,	/* putargr_d */
+    0,	/* putargi_d */
     8,	/* addr_d */
     26,	/* addi_d */
     12,	/* subr_d */
     26,	/* subi_d */
-    26,	/* rsbi_d */
+    30,	/* rsbi_d */
     8,	/* mulr_d */
     26,	/* muli_d */
     12,	/* divr_d */
@@ -313,7 +355,7 @@
     6,	/* str_d */
     10,	/* sti_d */
     7,	/* stxr_d */
-    7,	/* stxi_d */
+    8,	/* stxi_d */
     10,	/* bltr_d */
     28,	/* blti_d */
     10,	/* bler_d */
@@ -342,6 +384,11 @@
     28,	/* bordi_d */
     10,	/* bunordr_d */
     28,	/* bunordi_d */
+    0,	/* pushargr_d */
+    0,	/* pushargi_d */
+    0,	/* retr_d */
+    0,	/* reti_d */
+    10,	/* retval_d */
     0,	/* movr_w_f */
     0,	/* movr_ww_d */
     0,	/* movr_w_d */
@@ -351,23 +398,37 @@
     0,	/* movi_d_ww */
     0,	/* movr_d_w */
     0,	/* movi_d_w */
-    10,	/* x86_retval_f */
-    10,	/* x86_retval_d */
-#endif /* __X32 */
+#endif
 
 #if __X64
 #if __CYGWIN__
-#define JIT_INSTR_MAX 71
+#define JIT_INSTR_MAX 130
     0,	/* data */
     0,	/* live */
-    7,	/* align */
+    6,	/* align */
     0,	/* save */
     0,	/* load */
     0,	/* #name */
     0,	/* #note */
     7,	/* label */
-    71,	/* prolog */
+    130,	/* prolog */
+    0,	/* ellipsis */
+    0,	/* allocai */
+    0,	/* allocar */
     0,	/* arg */
+    0,	/* getarg_c */
+    0,	/* getarg_uc */
+    0,	/* getarg_s */
+    0,	/* getarg_us */
+    0,	/* getarg_i */
+    0,	/* getarg_ui */
+    0,	/* getarg_l */
+    0,	/* putargr */
+    0,	/* putargi */
+    7,	/* va_start */
+    7,	/* va_arg */
+    9,	/* va_arg_d */
+    0,	/* va_end */
     4,	/* addr */
     13,	/* addi */
     6,	/* addcr */
@@ -380,7 +441,7 @@
     13,	/* subci */
     9,	/* subxr */
     7,	/* subxi */
-    19,	/* rsbi */
+    16,	/* rsbi */
     7,	/* mulr */
     14,	/* muli */
     20,	/* qmulr */
@@ -483,7 +544,7 @@
     7,	/* stxr_c */
     7,	/* stxi_c */
     5,	/* stxr_s */
-    5,	/* stxi_s */
+    7,	/* stxi_s */
     4,	/* stxr_i */
     6,	/* stxi_i */
     4,	/* stxr_l */
@@ -532,13 +593,31 @@
     5,	/* jmpi */
     3,	/* callr */
     13,	/* calli */
-    68,	/* epilog */
+    0,	/* prepare */
+    0,	/* pushargr */
+    0,	/* pushargi */
+    0,	/* finishr */
+    0,	/* finishi */
+    0,	/* ret */
+    0,	/* retr */
+    0,	/* reti */
+    0,	/* retval_c */
+    0,	/* retval_uc */
+    0,	/* retval_s */
+    0,	/* retval_us */
+    0,	/* retval_i */
+    0,	/* retval_ui */
+    0,	/* retval_l */
+    124,	/* epilog */
     0,	/* arg_f */
+    0,	/* getarg_f */
+    0,	/* putargr_f */
+    0,	/* putargi_f */
     10,	/* addr_f */
     21,	/* addi_f */
     15,	/* subr_f */
     21,	/* subi_f */
-    21,	/* rsbi_f */
+    27,	/* rsbi_f */
     10,	/* mulr_f */
     21,	/* muli_f */
     15,	/* divr_f */
@@ -616,12 +695,20 @@
     25,	/* bordi_f */
     10,	/* bunordr_f */
     25,	/* bunordi_f */
+    0,	/* pushargr_f */
+    0,	/* pushargi_f */
+    0,	/* retr_f */
+    0,	/* reti_f */
+    0,	/* retval_f */
     0,	/* arg_d */
+    0,	/* getarg_d */
+    0,	/* putargr_d */
+    0,	/* putargi_d */
     10,	/* addr_d */
     25,	/* addi_d */
     15,	/* subr_d */
     25,	/* subi_d */
-    25,	/* rsbi_d */
+    27,	/* rsbi_d */
     10,	/* mulr_d */
     25,	/* muli_d */
     15,	/* divr_d */
@@ -699,6 +786,11 @@
     26,	/* bordi_d */
     11,	/* bunordr_d */
     26,	/* bunordi_d */
+    0,	/* pushargr_d */
+    0,	/* pushargi_d */
+    0,	/* retr_d */
+    0,	/* reti_d */
+    0,	/* retval_d */
     0,	/* movr_w_f */
     0,	/* movr_ww_d */
     0,	/* movr_w_d */
@@ -708,12 +800,10 @@
     0,	/* movi_d_ww */
     0,	/* movr_d_w */
     0,	/* movi_d_w */
-    0,	/* x86_retval_f */
-    0,	/* x86_retval_d */
 #else
 
 #  if __X64_32
-#define JIT_INSTR_MAX 44
+#define JIT_INSTR_MAX 108
     0,	/* data */
     0,	/* live */
     3,	/* align */
@@ -722,8 +812,24 @@
     0,	/* #name */
     0,	/* #note */
     3,	/* label */
-    39,	/* prolog */
+    108,	/* prolog */
+    0,	/* ellipsis */
+    0,	/* allocai */
+    0,	/* allocar */
     0,	/* arg */
+    0,	/* getarg_c */
+    0,	/* getarg_uc */
+    0,	/* getarg_s */
+    0,	/* getarg_us */
+    0,	/* getarg_i */
+    0,	/* getarg_ui */
+    0,	/* getarg_l */
+    0,	/* putargr */
+    0,	/* putargi */
+    41,	/* va_start */
+    45,	/* va_arg */
+    54,	/* va_arg_d */
+    0,	/* va_end */
     5,	/* addr */
     7,	/* addi */
     6,	/* addcr */
@@ -814,15 +920,15 @@
     0,	/* ldi_ui */
     0,	/* ldr_l */
     0,	/* ldi_l */
-    6,	/* ldxr_c */
+    9,	/* ldxr_c */
     7,	/* ldxi_c */
-    6,	/* ldxr_uc */
+    9,	/* ldxr_uc */
     7,	/* ldxi_uc */
-    6,	/* ldxr_s */
+    9,	/* ldxr_s */
     7,	/* ldxi_s */
-    6,	/* ldxr_us */
+    9,	/* ldxr_us */
     7,	/* ldxi_us */
-    5,	/* ldxr_i */
+    8,	/* ldxr_i */
     7,	/* ldxi_i */
     0,	/* ldxr_ui */
     0,	/* ldxi_ui */
@@ -836,11 +942,11 @@
     8,	/* sti_i */
     0,	/* str_l */
     0,	/* sti_l */
-    8,	/* stxr_c */
+    12,	/* stxr_c */
     7,	/* stxi_c */
-    6,	/* stxr_s */
+    10,	/* stxr_s */
     7,	/* stxi_s */
-    5,	/* stxr_i */
+    9,	/* stxr_i */
     6,	/* stxi_i */
     0,	/* stxr_l */
     0,	/* stxi_l */
@@ -888,8 +994,26 @@
     5,	/* jmpi */
     3,	/* callr */
     9,	/* calli */
+    0,	/* prepare */
+    0,	/* pushargr */
+    0,	/* pushargi */
+    0,	/* finishr */
+    0,	/* finishi */
+    0,	/* ret */
+    0,	/* retr */
+    0,	/* reti */
+    0,	/* retval_c */
+    0,	/* retval_uc */
+    0,	/* retval_s */
+    0,	/* retval_us */
+    0,	/* retval_i */
+    0,	/* retval_ui */
+    0,	/* retval_l */
     34,	/* epilog */
     0,	/* arg_f */
+    0,	/* getarg_f */
+    0,	/* putargr_f */
+    0,	/* putargi_f */
     10,	/* addr_f */
     21,	/* addi_f */
     15,	/* subr_f */
@@ -938,11 +1062,11 @@
     11,	/* movi_f */
     6,	/* ldr_f */
     10,	/* ldi_f */
-    7,	/* ldxr_f */
+    11,	/* ldxr_f */
     9,	/* ldxi_f */
     6,	/* str_f */
     10,	/* sti_f */
-    7,	/* stxr_f */
+    11,	/* stxr_f */
     9,	/* stxi_f */
     10,	/* bltr_f */
     21,	/* blti_f */
@@ -972,7 +1096,15 @@
     21,	/* bordi_f */
     10,	/* bunordr_f */
     21,	/* bunordi_f */
+    0,	/* pushargr_f */
+    0,	/* pushargi_f */
+    0,	/* retr_f */
+    0,	/* reti_f */
+    0,	/* retval_f */
     0,	/* arg_d */
+    0,	/* getarg_d */
+    0,	/* putargr_d */
+    0,	/* putargi_d */
     10,	/* addr_d */
     33,	/* addi_d */
     15,	/* subr_d */
@@ -1021,11 +1153,11 @@
     23,	/* movi_d */
     6,	/* ldr_d */
     10,	/* ldi_d */
-    7,	/* ldxr_d */
+    11,	/* ldxr_d */
     9,	/* ldxi_d */
     6,	/* str_d */
     10,	/* sti_d */
-    7,	/* stxr_d */
+    11,	/* stxr_d */
     9,	/* stxi_d */
     11,	/* bltr_d */
     34,	/* blti_d */
@@ -1055,6 +1187,11 @@
     34,	/* bordi_d */
     11,	/* bunordr_d */
     34,	/* bunordi_d */
+    0,	/* pushargr_d */
+    0,	/* pushargi_d */
+    0,	/* retr_d */
+    0,	/* reti_d */
+    0,	/* retval_d */
     0,	/* movr_w_f */
     0,	/* movr_ww_d */
     0,	/* movr_w_d */
@@ -1064,21 +1201,35 @@
     0,	/* movi_d_ww */
     0,	/* movr_d_w */
     0,	/* movi_d_w */
-    0,	/* x86_retval_f */
-    0,	/* x86_retval_d */
-#  else
 
-#define JIT_INSTR_MAX 43
+#  else
+#define JIT_INSTR_MAX 115
     0,	/* data */
     0,	/* live */
-    7,	/* align */
+    6,	/* align */
     0,	/* save */
     0,	/* load */
     0,	/* #name */
     0,	/* #note */
     7,	/* label */
-    43,	/* prolog */
+    115,	/* prolog */
+    0,	/* ellipsis */
+    0,	/* allocai */
+    0,	/* allocar */
     0,	/* arg */
+    0,	/* getarg_c */
+    0,	/* getarg_uc */
+    0,	/* getarg_s */
+    0,	/* getarg_us */
+    0,	/* getarg_i */
+    0,	/* getarg_ui */
+    0,	/* getarg_l */
+    0,	/* putargr */
+    0,	/* putargi */
+    42,	/* va_start */
+    41,	/* va_arg */
+    50,	/* va_arg_d */
+    0,	/* va_end */
     5,	/* addr */
     13,	/* addi */
     6,	/* addcr */
@@ -1091,7 +1242,7 @@
     13,	/* subci */
     9,	/* subxr */
     7,	/* subxi */
-    19,	/* rsbi */
+    16,	/* rsbi */
     7,	/* mulr */
     14,	/* muli */
     20,	/* qmulr */
@@ -1170,17 +1321,17 @@
     4,	/* ldr_l */
     8,	/* ldi_l */
     6,	/* ldxr_c */
-    5,	/* ldxi_c */
+    8,	/* ldxi_c */
     6,	/* ldxr_uc */
-    5,	/* ldxi_uc */
+    8,	/* ldxi_uc */
     6,	/* ldxr_s */
-    5,	/* ldxi_s */
+    8,	/* ldxi_s */
     6,	/* ldxr_us */
-    5,	/* ldxi_us */
+    8,	/* ldxi_us */
     5,	/* ldxr_i */
     7,	/* ldxi_i */
     5,	/* ldxr_ui */
-    4,	/* ldxi_ui */
+    6,	/* ldxi_ui */
     5,	/* ldxr_l */
     7,	/* ldxi_l */
     4,	/* str_c */
@@ -1192,9 +1343,9 @@
     4,	/* str_l */
     8,	/* sti_l */
     5,	/* stxr_c */
-    4,	/* stxi_c */
+    6,	/* stxi_c */
     6,	/* stxr_s */
-    5,	/* stxi_s */
+    7,	/* stxi_s */
     5,	/* stxr_i */
     6,	/* stxi_i */
     5,	/* stxr_l */
@@ -1239,22 +1390,40 @@
     10,	/* bxsubi */
     9,	/* bxsubr_u */
     10,	/* bxsubi_u */
-    0,	/* jmpr */
+    3,	/* jmpr */
     5,	/* jmpi */
     3,	/* callr */
     13,	/* calli */
+    0,	/* prepare */
+    0,	/* pushargr */
+    0,	/* pushargi */
+    0,	/* finishr */
+    0,	/* finishi */
+    0,	/* ret */
+    0,	/* retr */
+    0,	/* reti */
+    0,	/* retval_c */
+    0,	/* retval_uc */
+    0,	/* retval_s */
+    0,	/* retval_us */
+    0,	/* retval_i */
+    0,	/* retval_ui */
+    0,	/* retval_l */
     37,	/* epilog */
     0,	/* arg_f */
+    0,	/* getarg_f */
+    0,	/* putargr_f */
+    0,	/* putargi_f */
     10,	/* addr_f */
-    20,	/* addi_f */
+    21,	/* addi_f */
     15,	/* subr_f */
-    20,	/* subi_f */
-    20,	/* rsbi_f */
+    21,	/* subi_f */
+    30,	/* rsbi_f */
     10,	/* mulr_f */
-    20,	/* muli_f */
+    21,	/* muli_f */
     15,	/* divr_f */
-    20,	/* divi_f */
-    14,	/* negr_f */
+    21,	/* divi_f */
+    15,	/* negr_f */
     15,	/* absr_f */
     5,	/* sqrtr_f */
     11,	/* ltr_f */
@@ -1294,13 +1463,13 @@
     6,	/* ldr_f */
     10,	/* ldi_f */
     7,	/* ldxr_f */
-    6,	/* ldxi_f */
+    9,	/* ldxi_f */
     6,	/* str_f */
     10,	/* sti_f */
     7,	/* stxr_f */
-    7,	/* stxi_f */
+    9,	/* stxi_f */
     10,	/* bltr_f */
-    20,	/* blti_f */
+    21,	/* blti_f */
     10,	/* bler_f */
     25,	/* blei_f */
     12,	/* beqr_f */
@@ -1327,17 +1496,25 @@
     25,	/* bordi_f */
     10,	/* bunordr_f */
     25,	/* bunordi_f */
+    0,	/* pushargr_f */
+    0,	/* pushargi_f */
+    0,	/* retr_f */
+    0,	/* reti_f */
+    0,	/* retval_f */
     0,	/* arg_d */
+    0,	/* getarg_d */
+    0,	/* putargr_d */
+    0,	/* putargi_d */
     10,	/* addr_d */
     25,	/* addi_d */
     15,	/* subr_d */
     25,	/* subi_d */
-    25,	/* rsbi_d */
+    30,	/* rsbi_d */
     10,	/* mulr_d */
     25,	/* muli_d */
     15,	/* divr_d */
     25,	/* divi_d */
-    21,	/* negr_d */
+    22,	/* negr_d */
     16,	/* absr_d */
     5,	/* sqrtr_d */
     12,	/* ltr_d */
@@ -1377,11 +1554,11 @@
     6,	/* ldr_d */
     10,	/* ldi_d */
     7,	/* ldxr_d */
-    6,	/* ldxi_d */
+    9,	/* ldxi_d */
     6,	/* str_d */
     10,	/* sti_d */
     7,	/* stxr_d */
-    7,	/* stxi_d */
+    9,	/* stxi_d */
     11,	/* bltr_d */
     26,	/* blti_d */
     11,	/* bler_d */
@@ -1410,6 +1587,11 @@
     26,	/* bordi_d */
     11,	/* bunordr_d */
     26,	/* bunordi_d */
+    0,	/* pushargr_d */
+    0,	/* pushargi_d */
+    0,	/* retr_d */
+    0,	/* reti_d */
+    0,	/* retval_d */
     0,	/* movr_w_f */
     0,	/* movr_ww_d */
     0,	/* movr_w_d */
@@ -1419,8 +1601,6 @@
     0,	/* movi_d_ww */
     0,	/* movr_d_w */
     0,	/* movi_d_w */
-    0,	/* x86_retval_f */
-    0,	/* x86_retval_d */
 #endif /* __CYGWIN__ */
 #  endif /* __X64_32 */
 #endif /* __X64 */
