@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012, 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -58,6 +58,9 @@ static inline jit_node_t *_link_node(jit_state_t*, jit_node_t*);
 #define del_node(u, v)			_del_node(_jit, u, v)
 static inline void _del_node(jit_state_t*, jit_node_t*, jit_node_t*);
 
+#define free_node(u)			_free_node(_jit, u)
+static inline void _free_node(jit_state_t*, jit_node_t*);
+
 #define del_label(u, v)			_del_label(_jit, u, v)
 static void _del_label(jit_state_t*, jit_node_t*, jit_node_t*);
 
@@ -681,6 +684,8 @@ _new_node(jit_state_t *_jit, jit_code_t code)
 	new_pool();
     node = _jitc->list;
     _jitc->list = node->next;
+    if (_jitc->synth)
+	node->flag |= jit_flag_synth;
     node->next = NULL;
     node->code = code;
 
@@ -711,6 +716,14 @@ _del_node(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
     _jitc->list = node;
 }
 
+static inline void
+_free_node(jit_state_t *_jit, jit_node_t *node)
+{
+    memset(node, 0, sizeof(jit_node_t));
+    node->next = _jitc->list;
+    _jitc->list = node;
+}
+
 static void
 _del_label(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node)
 {
@@ -832,6 +845,13 @@ jit_new_state(void)
 void
 _jit_clear_state(jit_state_t *_jit)
 {
+#if DEVEL_DISASSEMBLER
+#  define jit_really_clear_state()	_jit_really_clear_state(_jit)
+}
+
+void _jit_really_clear_state(jit_state_t *_jit)
+{
+#endif
     jit_word_t		 offset;
     jit_function_t	*function;
 
@@ -889,6 +909,9 @@ _jit_clear_state(jit_state_t *_jit)
 void
 _jit_destroy_state(jit_state_t *_jit)
 {
+#if DEVEL_DISASSEMBLER
+    jit_really_clear_state();
+#endif
     if (!_jit->user_code)
 	munmap(_jit->code.ptr, _jit->code.length);
     if (!_jit->user_data)
@@ -896,6 +919,13 @@ _jit_destroy_state(jit_state_t *_jit)
     jit_free((jit_pointer_t *)&_jit);
 }
 
+void
+_jit_synth_inc(jit_state_t *_jit)
+{
+    assert(_jitc->synth < 8);
+    ++_jitc->synth;
+}
+
 jit_node_t *
 _jit_new_node(jit_state_t *_jit, jit_code_t code)
 {
@@ -917,6 +947,13 @@ _jit_link_node(jit_state_t *_jit, jit_node_t *node)
     link_node(node);
 }
 
+void
+_jit_synth_dec(jit_state_t *_jit)
+{
+    assert(_jitc->synth > 0);
+    --_jitc->synth;
+}
+
 jit_node_t *
 _jit_new_node_w(jit_state_t *_jit, jit_code_t code,
 		jit_word_t u)
@@ -927,6 +964,26 @@ _jit_new_node_w(jit_state_t *_jit, jit_code_t code,
     return (link_node(node));
 }
 
+jit_node_t *
+_jit_new_node_f(jit_state_t *_jit, jit_code_t code,
+		jit_float32_t u)
+{
+    jit_node_t		*node = new_node(code);
+    assert(!_jitc->realize);
+    node->u.f = u;
+    return (link_node(node));
+}
+
+jit_node_t *
+_jit_new_node_d(jit_state_t *_jit, jit_code_t code,
+		jit_float64_t u)
+{
+    jit_node_t		*node = new_node(code);
+    assert(!_jitc->realize);
+    node->u.d = u;
+    return (link_node(node));
+}
+
 jit_node_t *
 _jit_new_node_p(jit_state_t *_jit, jit_code_t code,
 		jit_pointer_t u)
@@ -955,6 +1012,28 @@ _jit_new_node_wp(jit_state_t *_jit, jit_code_t code,
     return (jit_new_node_ww(code, u, (jit_word_t)v));
 }
 
+jit_node_t *
+_jit_new_node_fp(jit_state_t *_jit, jit_code_t code,
+		 jit_float32_t u, jit_pointer_t v)
+{
+    jit_node_t		*node = new_node(code);
+    assert(!_jitc->realize);
+    node->u.f = u;
+    node->v.w = (jit_word_t)v;
+    return (link_node(node));
+}
+
+jit_node_t *
+_jit_new_node_dp(jit_state_t *_jit, jit_code_t code,
+		 jit_float64_t u, jit_pointer_t v)
+{
+    jit_node_t		*node = new_node(code);
+    assert(!_jitc->realize);
+    node->u.d = u;
+    node->v.w = (jit_word_t)v;
+    return (link_node(node));
+}
+
 jit_node_t *
 _jit_new_node_pw(jit_state_t *_jit, jit_code_t code,
 		 jit_pointer_t u, jit_word_t v)
@@ -1153,30 +1232,7 @@ _jit_prepare(jit_state_t *_jit)
     _jitc->function->call.argi =
 	_jitc->function->call.argf =
 	_jitc->function->call.size = 0;
-    _jitc->prepare = 1;
-}
-
-/* If declaring a jit function as varargs, in most backends it does
- * not change anything. Currently only exception is arm backend, that
- * if running in hardware float abi, switches to software float abi
- * if "self" function is varargs. Otherwise, there is no logic to
- * handle va_list like objects that need to parse runtime state, and
- * that is mainly because jit_arg* and jit_getarg* work only with
- * constants values, and one must not expect them to be handled at
- * runtime, they are parsed only once (same applies to jit_allocai,
- * that has no jit_allocar counterpart).
- */
-void
-_jit_ellipsis(jit_state_t *_jit)
-{
-    if (_jitc->prepare) {
-	assert(!(_jitc->function->call.call & jit_call_varargs));
-	_jitc->function->call.call |= jit_call_varargs;
-    }
-    else {
-	assert(!(_jitc->function->self.call & jit_call_varargs));
-	_jitc->function->self.call |= jit_call_varargs;
-    }
+    _jitc->prepare = jit_new_node(jit_code_prepare);
 }
 
 void
@@ -1195,27 +1251,69 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
     jit_int32_t		mask;
 
     switch (code) {
-	case jit_code_data:	case jit_code_align:	case jit_code_save:
-	case jit_code_load:	case jit_code_name:	case jit_code_label:
-	case jit_code_note:	case jit_code_prolog:	case jit_code_epilog:
+	case jit_code_data:	case jit_code_save:	case jit_code_load:
+	case jit_code_name:	case jit_code_label:	case jit_code_note:
+	case jit_code_prolog:	case jit_code_ellipsis:	case jit_code_epilog:
+	case jit_code_ret:	case jit_code_prepare:
 	    mask = 0;
 	    break;
-	case jit_code_live:
+	case jit_code_live:	case jit_code_va_end:
+	case jit_code_retr:	case jit_code_retr_f:	case jit_code_retr_d:
+	case jit_code_pushargr:	case jit_code_pushargr_f:
+	case jit_code_pushargr_d:
+	case jit_code_finishr:	/* synthesized will set jit_cc_a0_jmp */
 	    mask = jit_cc_a0_reg;
 	    break;
-	case jit_code_arg:	case jit_code_arg_f:	case jit_code_arg_d:
+	case jit_code_align:	case jit_code_reti:	case jit_code_pushargi:
+	case jit_code_finishi:	/* synthesized will set jit_cc_a0_jmp */
 	    mask = jit_cc_a0_int;
 	    break;
+	case jit_code_reti_f:	case jit_code_pushargi_f:
+	    mask = jit_cc_a0_flt;
+	    break;
+	case jit_code_reti_d:	case jit_code_pushargi_d:
+	    mask = jit_cc_a0_dbl;
+	    break;
+	case jit_code_allocai:
+	    mask = jit_cc_a0_int|jit_cc_a1_int;
+	    break;
+	case jit_code_arg:	case jit_code_arg_f:	case jit_code_arg_d:
+	    mask = jit_cc_a0_int|jit_cc_a0_arg;
+	    break;
 	case jit_code_calli:	case jit_code_jmpi:
 	    mask = jit_cc_a0_jmp;
 	    break;
 	case jit_code_callr:	case jit_code_jmpr:
 	    mask = jit_cc_a0_reg|jit_cc_a0_jmp;
 	    break;
-	case jit_code_x86_retval_f:
-	case jit_code_x86_retval_d:
+	case jit_code_retval_c:	case jit_code_retval_uc:
+	case jit_code_retval_s:	case jit_code_retval_us:
+	case jit_code_retval_i:	case jit_code_retval_ui:
+	case jit_code_retval_l:
+	case jit_code_retval_f:	case jit_code_retval_d:
+	case jit_code_va_start:
 	    mask = jit_cc_a0_reg|jit_cc_a0_chg;
 	    break;
+	case jit_code_getarg_c:	case jit_code_getarg_uc:
+	case jit_code_getarg_s:	case jit_code_getarg_us:
+	case jit_code_getarg_i:	case jit_code_getarg_ui:
+	case jit_code_getarg_l:
+	case jit_code_getarg_f:	case jit_code_getarg_d:
+	    mask = jit_cc_a0_reg|jit_cc_a0_chg|jit_cc_a1_arg;
+	    break;
+	case jit_code_putargr:	case jit_code_putargr_f:
+	case jit_code_putargr_d:
+	    mask = jit_cc_a0_reg|jit_cc_a1_arg;
+	    break;
+	case jit_code_putargi:
+	    mask = jit_cc_a0_int|jit_cc_a1_arg;
+	    break;
+	case jit_code_putargi_f:
+	    mask = jit_cc_a0_flt|jit_cc_a1_arg;
+	    break;
+	case jit_code_putargi_d:
+	    mask = jit_cc_a0_dbl|jit_cc_a1_arg;
+	    break;
 	case jit_code_movi:	case jit_code_ldi_c:	case jit_code_ldi_uc:
 	case jit_code_ldi_s:	case jit_code_ldi_us:	case jit_code_ldi_i:
 	case jit_code_ldi_ui:	case jit_code_ldi_l:	case jit_code_ldi_f:
@@ -1248,6 +1346,7 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
 	case jit_code_ldr_d:
 	case jit_code_movr_w_f:	case jit_code_movr_f_w:
 	case jit_code_movr_w_d:	case jit_code_movr_d_w:
+	case jit_code_va_arg:	case jit_code_va_arg_d:
 	    mask = jit_cc_a0_reg|jit_cc_a0_chg|jit_cc_a1_reg;
 	    break;
 	case jit_code_movr_d_ww:
@@ -1348,6 +1447,8 @@ _jit_classify(jit_state_t *_jit, jit_code_t code)
 	case jit_code_bordi_d:	case jit_code_bunordi_d:
 	    mask = jit_cc_a0_jmp|jit_cc_a1_reg|jit_cc_a2_dbl;
 	    break;
+	case jit_code_allocar:	/* synthesized instructions make it
+				 * equivalent to jit_cc_a0_chg */
 	case jit_code_str_c:	case jit_code_str_s:	case jit_code_str_i:
 	case jit_code_str_l:	case jit_code_str_f:	case jit_code_str_d:
 	    mask = jit_cc_a0_reg|jit_cc_a1_reg;
@@ -1459,8 +1560,10 @@ _jit_optimize(jit_state_t *_jit)
 	block = _jitc->blocks.ptr + offset;
 	if (!block->label)
 	    continue;
-	if (block->label->code != jit_code_epilog)
+	if (block->label->code != jit_code_epilog) {
 	    jit_setup(block);
+	    jit_regset_set(&block->setmask, &block->regmask);
+	}
     }
     /* call jit_update resolving undefined values in reverse
      * order so that sequential code would find most data already
@@ -1474,6 +1577,19 @@ _jit_optimize(jit_state_t *_jit)
 	    jit_update(block->label->next, &block->reglive, &_jitc->regmask);
 	}
     }
+    /* do a second pass from start to properly handle some conditions
+     * of very long living registers that are not referenced for
+     * several blocks */
+    bmp_zero();
+    for (offset = 0; offset < _jitc->blocks.offset; offset++) {
+	block = _jitc->blocks.ptr + offset;
+	if (!block->label)
+	    continue;
+	if (block->label->code != jit_code_epilog) {
+	    jit_regset_set(&_jitc->regmask, &block->setmask);
+	    jit_update(block->label->next, &block->reglive, &_jitc->regmask);
+	}
+    }
 
     patch_registers();
     simplify();
@@ -1501,6 +1617,13 @@ _jit_optimize(jit_state_t *_jit)
     }
 
     for (node = _jitc->head; node; node = node->next) {
+	mask = jit_classify(node->code);
+	if (mask & jit_cc_a0_reg)
+	    node->u.w &= ~jit_regno_patch;
+	if (mask & jit_cc_a1_reg)
+	    node->v.w &= ~jit_regno_patch;
+	if (mask & jit_cc_a2_reg)
+	    node->w.w &= ~jit_regno_patch;
 	switch (node->code) {
 	    case jit_code_prolog:
 		_jitc->function = _jitc->functions.ptr + node->w.w;
@@ -1515,9 +1638,16 @@ _jit_optimize(jit_state_t *_jit)
 		redundant_store(node, 0);
 		break;
 	    default:
-		mask = jit_classify(node->code);
 #if JIT_HASH_CONSTS
-		if (mask & jit_cc_a1_flt) {
+		if (mask & jit_cc_a0_flt) {
+		    node->u.p = jit_data(&node->u.f, sizeof(jit_float32_t), 4);
+		    node->flag |= jit_flag_node | jit_flag_data;
+		}
+		else if (mask & jit_cc_a0_dbl) {
+		    node->u.p = jit_data(&node->u.d, sizeof(jit_float64_t), 8);
+		    node->flag |= jit_flag_node | jit_flag_data;
+		}
+		else if (mask & jit_cc_a1_flt) {
 		    node->v.p = jit_data(&node->v.f, sizeof(jit_float32_t), 4);
 		    node->flag |= jit_flag_node | jit_flag_data;
 		}
@@ -1551,6 +1681,10 @@ _jit_optimize(jit_state_t *_jit)
 			(jit_cc_a1_reg|jit_cc_a1_chg))
 			jit_regset_setbit(&_jitc->function->regset,
 					  jit_regno(node->v.w));
+		    if ((mask & (jit_cc_a2_reg|jit_cc_a2_chg)) ==
+			(jit_cc_a2_reg|jit_cc_a2_chg))
+			jit_regset_setbit(&_jitc->function->regset,
+					  jit_regno(node->w.w));
 		}
 		break;
 	}
@@ -1582,20 +1716,6 @@ _jit_reglive(jit_state_t *_jit, jit_node_t *node)
 		else if (!(spec & jit_class_sav))
 		    jit_regset_clrbit(&_jitc->reglive, value);
 	    }
-#if defined(JIT_RET)
-	    /* Explicitly set return registers as live, because retval
-	     * should be free to not create a note, and/or user not
-	     * call jit_retval (but not a good idea to expect JIT_R0
-	     * to match JIT_RET) */
-	    jit_regset_setbit(&_jitc->reglive, JIT_RET);
-#  if __arm__
-	    /* FIXME need a better logic (and r2-r3 may contain results) */
-	    jit_regset_setbit(&_jitc->reglive, _R1);
-#  endif
-#endif
-#if defined(JIT_FRET)
-	    jit_regset_setbit(&_jitc->reglive, JIT_FRET);
-#endif
 	    break;
 	default:
 	    value = jit_classify(node->code);
@@ -1637,8 +1757,14 @@ _jit_reglive(jit_state_t *_jit, jit_node_t *node)
 		else
 		    jit_regset_setbit(&_jitc->reglive, node->v.w);
 	    }
-	    if ((value & jit_cc_a2_reg) && !(node->w.w & jit_regno_patch))
-		jit_regset_setbit(&_jitc->reglive, node->w.w);
+	    if ((value & jit_cc_a2_reg) && !(node->w.w & jit_regno_patch)) {
+		if (value & jit_cc_a2_chg) {
+		    jit_regset_clrbit(&_jitc->reglive, node->w.w);
+		    jit_regset_setbit(&_jitc->regmask, node->w.w);
+		}
+		else
+		    jit_regset_setbit(&_jitc->reglive, node->w.w);
+	    }
 	    if (jit_regset_set_p(&_jitc->regmask)) {
 		bmp_zero();
 		jit_update(node->next, &_jitc->reglive, &_jitc->regmask);
@@ -2044,8 +2170,11 @@ _jit_setup(jit_state_t *_jit, jit_block_t *block)
 		if ((value & jit_cc_a2_reg) &&
 		    !(node->w.w & jit_regno_patch) &&
 		    jit_regset_tstbit(&regmask, node->w.w)) {
-		    jit_regset_clrbit(&regmask, node->w.w);
-		    jit_regset_setbit(&reglive, node->w.w);
+		    live = !(value & jit_cc_a2_chg);
+		    if (live || !jump)
+			jit_regset_clrbit(&regmask, node->w.w);
+		    if (live)
+			jit_regset_setbit(&reglive, node->w.w);
 		}
 		if (value & jit_cc_a0_jmp)
 		    jump = 1;
@@ -2081,36 +2210,21 @@ _jit_update(jit_state_t *_jit, jit_node_t *node,
 	switch (node->code) {
 	    case jit_code_label:
 		block = _jitc->blocks.ptr + node->v.w;
-		if (bmp_tst(node->v.w))
-		    return;
-		bmp_set(node->v.w);
 		jit_regset_and(&ztmp, mask, &block->reglive);
 		if (jit_regset_set_p(&ztmp)) {
 		    jit_regset_ior(live, live, &ztmp);
 		    jit_regset_com(&ztmp, &ztmp);
 		    jit_regset_and(mask, mask, &ztmp);
 		}
+		if (bmp_tst(node->v.w))
+		    return;
+		bmp_set(node->v.w);
 		break;
 	    case jit_code_prolog:
 		jit_regset_set_ui(mask, 0);
 		return;
 	    case jit_code_epilog:
 		jit_regset_set_ui(mask, 0);
-#if defined(JIT_RET)
-		/* On some backends it may be required to allocate one
-		 * or more registers to jump from a jit_ret* to the
-		 * epilog label.
-		 * Because currently there is no type information,
-		 * assume JIT_RET and JIT_FRET are live in the epilog.
-		 * Only JIT_RET really should be marked as live, to
-		 * prevent it being allocated, usually in the jump to
-		 * the epilog, but also mark JIT_FRET as live for the
-		 * sake of correctness. */
-		jit_regset_setbit(live, JIT_RET);
-#endif
-#if defined(JIT_FRET)
-		jit_regset_setbit(live, JIT_FRET);
-#endif
 		return;
 	    case jit_code_callr:
 		value = jit_regno(node->u.w);
@@ -2121,24 +2235,6 @@ _jit_update(jit_state_t *_jit, jit_node_t *node,
 		    }
 		}
 	    case jit_code_calli:
-#if defined(JIT_RET)
-		if (jit_regset_tstbit(mask, JIT_RET)) {
-		    jit_regset_setbit(live, JIT_RET);
-		    jit_regset_clrbit(mask, JIT_RET);
-		}
-#  if __arm__
-		if (jit_regset_tstbit(mask, _R1)) {
-		    jit_regset_setbit(live, _R1);
-		    jit_regset_clrbit(mask, _R1);
-		}
-#  endif
-#endif
-#if defined(JIT_FRET)
-		if (jit_regset_tstbit(mask, JIT_FRET)) {
-		    jit_regset_setbit(live, JIT_FRET);
-		    jit_regset_clrbit(mask, JIT_FRET);
-		}
-#endif
 		for (value = 0; value < _jitc->reglen; ++value) {
 		    value = jit_regset_scan1(mask, value);
 		    if (value >= _jitc->reglen)
@@ -2156,7 +2252,8 @@ _jit_update(jit_state_t *_jit, jit_node_t *node,
 		    if (!(node->w.w & jit_regno_patch)) {
 			if (jit_regset_tstbit(mask, node->w.w)) {
 			    jit_regset_clrbit(mask, node->w.w);
-			    jit_regset_setbit(live, node->w.w);
+			    if (!(value & jit_cc_a2_chg))
+				jit_regset_setbit(live, node->w.w);
 			}
 		    }
 		}
@@ -2659,6 +2756,11 @@ _redundant_store(jit_state_t *_jit, jit_node_t *node, jit_bool_t jump)
 		    if (regno == jit_regno(iter->v.w))
 			return;
 		}
+		if ((spec & (jit_cc_a2_reg|jit_cc_a2_chg)) ==
+		    (jit_cc_a2_reg|jit_cc_a2_chg)) {
+		    if (regno == jit_regno(iter->w.w))
+			return;
+		}
 		break;
 	}
     }
@@ -2688,7 +2790,7 @@ _simplify_movr(jit_state_t *_jit, jit_node_t *prev, jit_node_t *node,
     else {
 	value->kind = jit_kind_register;
 	value->base.q.l = right;
-	value->base.q.h = _jitc->gen[regno];
+	value->base.q.h = _jitc->gen[right];
     }
     ++_jitc->gen[regno];
 
@@ -2964,6 +3066,11 @@ _simplify(jit_state_t *_jit)
 		    _jitc->values[regno].kind = 0;
 		    ++_jitc->gen[regno];
 		}
+		if (info & jit_cc_a2_chg) {
+		    regno = jit_regno(node->w.w);
+		    _jitc->values[regno].kind = 0;
+		    ++_jitc->gen[regno];
+		}
 		break;
 	}
     }
@@ -2999,6 +3106,9 @@ _register_change_p(jit_state_t *_jit, jit_node_t *node, jit_node_t *link,
 		else if ((value & jit_cc_a1_reg) && node->v.w == regno &&
 			 (value & jit_cc_a1_chg))
 		    return (jit_reg_change);
+		else if ((value & jit_cc_a2_reg) && node->w.w == regno &&
+			 (value & jit_cc_a2_chg))
+		    return (jit_reg_change);
 	}
     }
 
