@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013  Free Software Foundation, Inc.
+ * Copyright (C) 2013-2017  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -22,7 +22,7 @@
 #include <stdio.h>
 #include "lib/jit_names.c"
 
-jit_int16_t	_szs[jit_code_last_code + 1];
+jit_int16_t	_szs[jit_code_last_code];
 
 int
 main(int argc, char *argv[])
@@ -40,7 +40,7 @@ main(int argc, char *argv[])
     fclose(fp);
 
     max = 0;
-    for (offset = 0; offset <= jit_code_last_code; offset++)
+    for (offset = 0; offset < jit_code_last_code; offset++)
 	if (max < _szs[offset])
 	    max = _szs[offset];
 
@@ -80,9 +80,11 @@ main(int argc, char *argv[])
     fprintf(fp, "#if defined(__ppc__)\n");
 #elif defined(__powerpc__)
     fprintf(fp, "#if defined(__powerpc__)\n");
+    fprintf(fp, "#if __BYTE_ORDER == %s\n",
+	    __BYTE_ORDER == __BIG_ENDIAN ? "__BIG_ENDIAN" : "__LITTLE_ENDIAN");
 #endif
     fprintf(fp, "#define JIT_INSTR_MAX %d\n", max);
-    for (offset = 0; offset <= jit_code_last_code; offset++)
+    for (offset = 0; offset < jit_code_last_code; offset++)
 	fprintf(fp, "    %d,	/* %s */\n", _szs[offset], code_name[offset]);
 #if defined(__arm__)
     fprintf(fp, "#endif /* __ARM_PCS_VFP */\n");
@@ -93,6 +95,7 @@ main(int argc, char *argv[])
 #elif defined(__ppc__)
     fprintf(fp, "#endif /* __ppc__ */\n");
 #elif defined(__powerpc__)
+    fprintf(fp, "#endif /* __BYTE_ORDER */\n");
     fprintf(fp, "#endif /* __powerpc__ */\n");
 #endif
 #if __X64 || __X32
@@ -100,7 +103,7 @@ main(int argc, char *argv[])
     fprintf(fp, "#  endif /* __X64_32 */\n");
     fprintf(fp, "#endif /* __X64 */\n");
 #  else
-    fprintf(fp, "#if __X32\n");
+    fprintf(fp, "#endif /* __X32 */\n");
 #  endif
 #else
     fprintf(fp, "#endif /* __WORDSIZE */\n");
